Require Import ssreflect ssrfun ssrbool ssrnat seq eqtype.
Require Import prelude heaps terms prefix xfindCTC.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* NOT WORKING VERSION WITH OUTPUTS AS FIELDS *)

Definition invariant i j t h := [/\ interp j t = h, subctx i j & valid j t].

(* Main structure
   i : input context
   j : output context
   t : syntactification of heap_of using context j *) 
Class Ast (i : ctx)  (h : heap) :=
       { ctx_of : ctx;
         term_of : synheap;
         ast : invariant i ctx_of term_of h}.

Implicit Arguments Ast [].
Implicit Arguments ctx_of [i h].
Implicit Arguments term_of [i h].


(* pass output context of f1 as input of f2 *)
Program Instance
  union_struct i h1 h2 (f1 : Ast i h1) (f2 : Ast (ctx_of f1) h2) : 
  Ast i (h1 :+ h2) | 3 := {| ctx_of := (ctx_of f2); term_of := (term_of f1) ++ (term_of f2) |}.
Next Obligation.
case: f1 f2=>[j t1 /= [<- S1 D1]] [k t2 /= [<- S2 D2]].
split; first by rewrite interp_cat (interp_subctx D1 S2).
- by apply: (subctx_trans S1 S2).
by rewrite valid_cat D2 andbT; apply: (valid_subctx S2).
Qed.

Program Instance empty_struct i : Ast i empty | 1 := 
  {| ctx_of := i; term_of := [::] |}.
Next Obligation.
split; by [|apply: subctx_refl|].
Qed.

Program Instance
  pts_struct A hs xs1 x (d : A)
           (f : XFind xs1 x) :
  Ast (Context hs xs1) (x :-> d) | 2 := 
  {| ctx_of := (Context hs seq_of); 
     term_of := [:: Pts index_of (dyn d)] |}.
Next Obligation.
case: f=>[xs2 n /= [H P]]; split; first by rewrite /= H.
- by split; [apply: prefix_refl|].
by apply/andP; rewrite /= (onth_size H). 
Qed.

Program Instance var_struct hs1 xs h (f : XFind hs1 h) :
  Ast (Context hs1 xs) h | 1000 :=
  {| ctx_of := (Context seq_of xs); 
     term_of := [:: Var index_of] |}. 
Next Obligation.
case:f=>hs2 n [H1 H2]; split; first by rewrite /= /hlook H1.
- by split; [|apply: prefix_refl].
by apply/andP; rewrite /= (onth_size H1).
Qed.

(* The main lemma *)
Theorem cancelR h1 h2 (f1 : Ast empc h1) (f2 : Ast (ctx_of f1) h2) :
        h1 = h2 -> 
        def h1 ->
        let k := ctx_of f2 in
        eval k (cancel k (term_of f1) (term_of f2)).
Proof.
case: f1 f2=>[j t1 /= [<- _ I]] [k t2 /= [<- S _]] D H.
by apply: cancel_sound; rewrite -(interp_subctx I S).
Qed.

Implicit Arguments cancelR [h1 h2 f1 f2].

Hint Extern 5 (Ast _ _) => progress simpl : typeclass_instances.
(*Hint Extern 0 (term_of _ _ _) => progress simpl : typeclass_instances.
*)
Lemma refl h1 (f : Ast empc h1) : def h1 -> 
  interp (ctx_of f) (term_of f) = h1. 
Proof. by case: f=>[i t /= [->]]. Qed.

(************)
(* Examples *)  
(************)
Example ex1 x (h:heap) (v1 v2:nat): 
          def (x :-> v1) -> x :-> v1 = x :-> v2 -> 
          if v1 == v2 then true else false.
move=>D H.
move: (cancelR H D)=>/= {H}H.
by move/dyn_inj: H=>->; rewrite eq_refl.
Abort.

Example ex2 h1 h2 h3 h4 x1 x2 (d1 d2 d3 d4 : nat) : 
     def ((h3 :+ (x1 :-> d1)) :+ (h1 :+ empty) :+ (x2 :-> d2)) ->
     (h3 :+ (x1 :-> d1)) :+ (h1 :+ empty) :+ (x2 :-> d2) = 
     (x2 :-> d3) :+ (h2 :+ empty :+ h3) :+ h4 :+ (x1 :-> d4) ->
     d1 = d4 /\ d2 = d3 /\ h1 = h2 :+ h4.
move=>D H.
eapply refl in D. 
Abort.