Require Import ssreflect ssrbool ssrnat eqtype seq ssrfun.
Require Import heaps rels hprop stmod stsep stlog. 
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.


Class Update (h1 h2 k1 k2 : heap) :=
  { rest : heap; update1 : h1 = k1 :+ rest; update2 : h2 = k2 :+ rest}.

Implicit Arguments update1 [h1 h2 k1 k2].
Implicit Arguments update2 [h1 h2 k1 k2].
Implicit Arguments rest [h1 h2 k1 k2].


Program Instance found_struct k1 k2 : Update k1 k2 k1 k2 | 1 := {| rest := empty |}.
Next Obligation. by rewrite unh0. Qed.
Next Obligation. by rewrite unh0. Qed.

Program Instance left_struct l h1 h2 k1 k2 (f : Update h1 h2 k1 k2) :
  Update (l :+ h1) (l :+ h2) k1 k2 | 2 := {| rest := (l :+ rest f) |}.
Next Obligation. 
have H : h1 = k1 :+ (rest f) by eapply (update1 f).
by rewrite -unCA -H.
Qed.
Next Obligation.
have H : h2 = k2 :+ (rest f) by eapply (update2 f).
by rewrite -unCA -H.
Qed.

Program Instance right_struct l h1 h2 k1 k2 (f : Update h1 h2 k1 k2) :
  Update (h1 :+ l) (h2 :+ l) k1 k2 | 2 := {| rest := (rest f :+ l) |}.
Next Obligation. 
have H : h1 = k1 :+ (rest f) by eapply (update1 f).
by rewrite unA -H.
Qed.
Next Obligation.
have H : h2 = k2 :+ (rest f) by eapply (update2 f).
by rewrite unA -H.
Qed.

Notation cont A := (ans A -> heap -> Prop).


Section EvalWriteR. 
Variables (A B C : Type).
(*
Lemma val_writeR h1 h2 (v : A) (w : B) x i (f1 : Update h1 (x:->v) i) (f2 : Update h2 (x:->w) i) (r : cont unit) : 
        (def h1 -> r (Val tt) h1) -> 
        verify h2 (write_s x v) r.
Proof. by rewrite (update _ f1) (update _ f2); apply: val_write. Qed.

Lemma try_writeR s1 s2 (v : A) (w : C) x h i 
                 (f : Update h k i) (r : cont B) : 
        verify (f (x :-> v)) (s1 tt) r -> 
        verify (f (x :-> w)) (try_s (write_s x v) s1 s2) r. 
Proof. rewrite !updateE; apply: try_write. Qed.
*)

Lemma bnd_writeR s (v : A) (w : C) x h1 h2 (f : Update h1 h2 (x:->v) (x:->w)) (r : cont B) : 
        verify (s tt) h1 r -> 
        verify (bind_s (write_s x v) s) h2 r. 
Proof. 
set l := rest f.
have H1 : h1 = (x :-> v) :+ l by eapply (update1 f).
have H2 : h2 = (x :-> w) :+ l by eapply (update2 f).
by rewrite H1 H2; apply: bnd_write. 
Qed.

End EvalWriteR.

Example ex_bwd i x1 x2 (e : unit -> spec nat) q: 
          verify (e tt) (i :+ (x1 :-> 1 :+ x2 :-> 4)) q ->
          verify (bind_s (write_s x2 4) e) (i :+ (x1 :-> 1 :+ x2 :-> 2)) q.
move=>H.
by apply: bnd_writeR.
Abort.


Example ex_fwd i x1 x2 (e : unit -> spec nat) q: 
          verify (e tt) (i :+ (x1 :-> 1 :+ x2 :-> 4)) q ->
          verify (bind_s (write_s x2 4) e) (i :+ (x1 :-> 1 :+ x2 :-> 2)) q.
move=>H. 
by apply: (bnd_writeR _ H).
Abort.


Example ex_stress A (x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 : ptr)
              (v1 v2 v3 v4 v5 v6 v7 v8 v9 v10 : A) 
              (h1 h2 h3 h4 h5 h6 h7 h8 h9 h10 : heap) (e : unit -> spec nat) q: 
          verify (e tt) (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v9 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) q ->
          verify (bind_s (write_s x7 v9) e) (h1 :+ h2 :+ h3 :+ h4 :+ h5 :+ 
               x1 :-> v1 :+ x2 :-> v2 :+ x3 :-> v3 :+ x4 :-> v4 :+ x5 :-> v5 :+ 
               x6 :-> v6 :+ x7 :-> v7 :+ x8 :-> v8 :+ x9 :-> v9 :+ x10 :-> v10 :+ 
               h6 :+ h7 :+ h8 :+ h9 :+ h10) q.
move=>H.
Time by apply: bnd_writeR. (* 0s *)
Time Qed.
