Set Automatic Coercions Import.

Require Import ssreflect ssrbool ssrnat eqtype seq ssrfun.
Require Import heaps rels hprop stmod stsep2 stlog2. 
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* EXAMPLE MATCHING THE LAST PARAMETER OF THE LEMMA *)

(*************************************************************************)
(* First, the mechanism for search-and-replace for the overloaded lemas, *)
(* pattern-matching on heap expressions.                                 *)
(*************************************************************************)

Structure tagged_heap := Tag {untag :> heap}.

Definition right_tag := Tag.
Definition left_tag := right_tag.
Canonical Structure found_tag i := left_tag i.

Definition update_axiom k r (h : tagged_heap) := untag h = k :+ r.

Structure update (k r : heap) :=
  Update {heap_of :> tagged_heap; 
        _ : update_axiom k r heap_of}.

Lemma updateE r k (f : update k r) : untag f = k :+ r.
Proof. by case: f=>[[j]] /=; rewrite /update_axiom /= => ->. Qed.

Lemma found_pf k : update_axiom k empty (found_tag k). 
Proof. by rewrite /update_axiom unh0. Qed.
 
Canonical Structure found_struct k := Update (found_pf k).

Lemma left_pf h r (f : forall k, update k r) k : 
        update_axiom k (r :+ h) (left_tag (f k :+ h)).
Proof. by rewrite updateE /update_axiom /= unA. Qed.

Canonical Structure left_struct h r (f : forall k, update k r) k :=
  Update (left_pf h f k).

Lemma right_pf h r (f : forall k, update k r) k :
        update_axiom k (h :+ r) (right_tag (h :+ f k)).
Proof. by rewrite updateE /update_axiom /= unCA. Qed.

Canonical Structure right_struct h r (f : forall k, update k r) k := 
  Update (right_pf h f k).

Section EvalWriteR. 
Variables (A B C : Type).

Notation cont A := (ans A -> heap -> Prop).

(* Structure keying on the last parameter of the verify predicate *)
Structure verify_write x (v : A) (w : C) h1 h2  := VW {
  q_of : cont B;
  proof s : verify h1 (s tt) q_of -> 
       verify h2 (bind_s (write_s x v) s) q_of 
}.

Program
Canonical Structure solve_vw x (v : A) (w : C) i (f : forall k, update k i) q :=
  @VW x v w (f (x :-> v)) (f (x :-> w)) q _.
Next Obligation. move: H; rewrite !updateE. apply: bnd_write. Qed.


Lemma bnd_writeR s (v : A) (w : C) x h1 h2 (f : verify_write x v w h1 h2) : 
        verify h1 (s tt) (q_of f) -> 
        verify h2 (bind_s (write_s x v) s) (q_of f). 
Proof. by apply: proof. Qed.

End EvalWriteR.


Example ex_bwd i x1 x2 (e : unit -> spec nat) q: 
          verify (i :+ (x1 :-> 1 :+ x2 :-> 4)) (e tt) q ->
          verify (i :+ (x1 :-> 1 :+ x2 :-> 2)) (bind_s (write_s x2 4) e) q.
by move=>H; apply: bnd_writeR.
Abort.


Example ex_fwd i x1 x2 (e : unit -> spec nat) q: 
          verify (i :+ (x1 :-> 1 :+ x2 :-> 4)) (e tt) q ->
          verify (i :+ (x1 :-> 1 :+ x2 :-> 2)) (bind_s (write_s x2 4) e) q.
move=>H. 
by move: (bnd_writeR (x:=x2) (w:=2) H).
Abort.
