Require Import ssreflect ssrfun ssrbool ssrnat seq prefix.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(******************************************************************************)
(*       Module for searching and inserting elements in a list                *)
(******************************************************************************)

Definition invariant A s r i (e : A) := onth r i = Some e /\ prefix s r.

Class XFind A (s : seq A) (e : A) := {
  seq_of : seq A;
  index_of : nat;
  xfind : invariant s seq_of index_of e}.

Implicit Arguments XFind [A].

Program Instance found_struct A (x:A) t : XFind (x :: t) x := {| seq_of := (x :: t); index_of := 0|}.
Next Obligation. by split; [|apply: prefix_refl]. Qed.

Program Instance recurse_struct A (y:A) t e (f : XFind t e) :
  XFind (y :: t) e | 2 := {| seq_of := (y :: seq_of); index_of := index_of.+1|}.
Next Obligation.
by case:f=>r i /= [[H1 H2]]; split; [|apply/prefix_cons].
Qed.

Program Instance extend_struct A (x:A) : XFind [::] x := {| seq_of := [:: x]; index_of := 0|}.
Next Obligation. by []. Qed.

Hint Extern 1 (XFind _ _) => progress simpl : typeclass_instances.

Example unit_test A (x1 x2 x3 x y : A):
   (forall s e (f : XFind s e), nth x1 seq_of index_of = e -> e = x) ->
  x = x.
Proof.
move=>test_form.
apply: (test_form [::]). simpl.
apply: (test_form [:: x1; x]). simpl.
apply: (test_form [:: x1; x2; x; x3]). simpl.
apply: (test_form [:: x1; x2; x3]). rewrite [seq_of]/=. rewrite [index_of]/=. simpl.
Abort.

(*
Lemma two (n m : nat) (f1 : XFind nil n) (f2 : XFind (@seq_of _ _ _ f1) m) : 
  size (@seq_of _ _ _ f2) > 1 -> n <> m.
Admitted.

Example ex : 1 <> 3.
Set Printing All.
apply: two.
simpl.
Abort.
*)