Require Import Coq.Lists.List.
Set Implicit Arguments.
Unset Strict Implicit.

(******************************************************************************)
(*       Module for searching and inserting elements in a list                *)
(******************************************************************************)

Class XFind A (s : list A) (e : A) := {
  seq_of : list A;
  index_of : nat;
  xfind : nth index_of seq_of e = e}.

Implicit Arguments XFind [A].

Program Instance found_struct A (x:A) t : XFind (x :: t) x := 
  {| seq_of := (x :: t); index_of := 0|}.

Program Instance recurse_struct A (y:A) t e (f : XFind t e) :
  XFind (y :: t) e | 100 := {| seq_of := (y :: seq_of); index_of := (S index_of)|}.
Next Obligation.
case f; intros r i; auto.
Qed.

Program Instance extend_struct A (x:A) : XFind nil x := 
  {| seq_of := (x :: nil); index_of := 0|}.

Example unit_test A (x1 x2 x3 x y : A):
   (forall s e (f : XFind s e), nth index_of seq_of x1 = e -> e = x) ->
  x = x.
Proof.
intro test_form.
eapply (test_form nil). simpl.
eapply (test_form (x1 :: x :: nil)). simpl.
eapply (test_form (x1 :: x2 :: x :: x3 :: nil)). simpl.
eapply (test_form (x1 :: x2 :: x3 :: nil)). simpl.
Abort.

Hint Extern 4 (XFind _ _) => progress simpl : typeclass_instances.

Lemma two (n m : nat) (f1 : XFind nil n) (f2 : XFind (@seq_of _ _ _ f1) m) : 
  length (@seq_of _ _ _ f2) > 1 -> n <> m.
Admitted.

Example ex : 1 <> 3.
Set Printing All.
eapply two.
simpl.
Abort.